@@ -5,7 +5,7 @@ module Development.Shake.Internal.Options(
     Progress(..), Verbosity(..), Rebuild(..), Lint(..), Change(..),
     ShakeOptions(..), shakeOptions,
     -- Internal stuff
-    shakeRebuildApply, shakeAbbreviationsApply
+    shakeRebuildApply, shakeAbbreviationsApply, shakeOptionsFields
     ) where
 
 import Data.Data
@@ -13,12 +13,17 @@ import Data.List.Extra
 import Data.Tuple.Extra
 import Data.Maybe
 import Data.Dynamic
+import Control.Monad
+import General.Extra
+import System.Time.Extra
 import qualified Data.HashMap.Strict as Map
-import Development.Shake.Internal.Progress
 import Development.Shake.Internal.FilePattern
+import Development.Shake.Internal.Errors
 import qualified Data.ByteString.Char8 as BS
 import qualified Data.ByteString.UTF8 as UTF8
 import Development.Shake.Internal.CmdOption
+import Data.Semigroup
+import Prelude
 
 
 -- | The current assumptions made by the build system, used by 'shakeRebuild'. These options
@@ -82,6 +87,46 @@ data Change
       deriving (Eq,Ord,Show,Read,Typeable,Data,Enum,Bounded)
 
 
+-- | Information about the current state of the build, obtained by either passing a callback function
+--   to 'Development.Shake.shakeProgress' (asynchronous output) or 'Development.Shake.getProgress'
+--   (synchronous output). Typically a build system will pass 'progressDisplay' to 'Development.Shake.shakeProgress',
+--   which will poll this value and produce status messages.
+data Progress = Progress
+-- In retrospect shakeProgress should have been done differently, as a feature you turn on in Rules
+-- but easiest way around that for now is put the Progress type in Options
+
+    {isFailure :: !(Maybe String) -- ^ Starts out 'Nothing', becomes 'Just' a target name if a rule fails.
+    ,countSkipped :: {-# UNPACK #-} !Int -- ^ Number of rules which were required, but were already in a valid state.
+    ,countBuilt :: {-# UNPACK #-} !Int -- ^ Number of rules which were have been built in this run.
+    ,countUnknown :: {-# UNPACK #-} !Int -- ^ Number of rules which have been built previously, but are not yet known to be required.
+    ,countTodo :: {-# UNPACK #-} !Int -- ^ Number of rules which are currently required (ignoring dependencies that do not change), but not built.
+    ,timeSkipped :: {-# UNPACK #-} !Double -- ^ Time spent building 'countSkipped' rules in previous runs.
+    ,timeBuilt :: {-# UNPACK #-} !Double -- ^ Time spent building 'countBuilt' rules.
+    ,timeUnknown :: {-# UNPACK #-} !Double -- ^ Time spent building 'countUnknown' rules in previous runs.
+    ,timeTodo :: {-# UNPACK #-} !(Double,Int) -- ^ Time spent building 'countTodo' rules in previous runs, plus the number which have no known time (have never been built before).
+    }
+    deriving (Eq,Ord,Show,Read,Data,Typeable)
+
+instance Semigroup Progress where
+    a <> b = Progress
+        {isFailure = isFailure a `mplus` isFailure b
+        ,countSkipped = countSkipped a + countSkipped b
+        ,countBuilt = countBuilt a + countBuilt b
+        ,countUnknown = countUnknown a + countUnknown b
+        ,countTodo = countTodo a + countTodo b
+        ,timeSkipped = timeSkipped a + timeSkipped b
+        ,timeBuilt = timeBuilt a + timeBuilt b
+        ,timeUnknown = timeUnknown a + timeUnknown b
+        ,timeTodo = let (a1,a2) = timeTodo a; (b1,b2) = timeTodo b
+                        x1 = a1 + b1; x2 = a2 + b2
+                    in x1 `seq` x2 `seq` (x1,x2)
+        }
+
+instance Monoid Progress where
+    mempty = Progress Nothing 0 0 0 0 0 0 0 (0,0)
+    mappend = (<>)
+
+
 -- | Options to control the execution of Shake, usually specified by overriding fields in
 --   'shakeOptions':
 --
@@ -94,6 +139,7 @@ data ShakeOptions = ShakeOptions
         -- ^ Defaults to @.shake@. The directory used for storing Shake metadata files.
         --   All metadata files will be named @'shakeFiles'\/.shake./file-name/@, for some @/file-name/@.
         --   If the 'shakeFiles' directory does not exist it will be created.
+        --   If set to @\"\/dev\/null\"@ then no shakeFiles are read or written (even on Windows).
     ,shakeThreads :: Int
         -- ^ Defaults to @1@. Maximum number of rules to run in parallel, similar to @make --jobs=/N/@.
         --   For many build systems, a number equal to or slightly less than the number of physical processors
@@ -105,7 +151,7 @@ data ShakeOptions = ShakeOptions
         --   significant changes to the rules that require a wipe. The version number should be
         --   set in the source code, and not passed on the command line.
     ,shakeVerbosity :: Verbosity
-        -- ^ Defaults to 'Normal'. What level of messages should be printed out.
+        -- ^ Defaults to 'Info'. What level of messages should be printed out.
     ,shakeStaunch :: Bool
         -- ^ Defaults to 'False'. Operate in staunch mode, where building continues even after errors,
         --   similar to @make --keep-going@.
@@ -121,9 +167,11 @@ data ShakeOptions = ShakeOptions
         -- ^ Directories in which the files will be tracked by the linter.
     ,shakeLintIgnore :: [FilePattern]
         -- ^ File patterns which are ignored from linter tracking, a bit like calling 'Development.Shake.trackAllow' in every rule.
+    ,shakeLintWatch :: [FilePattern]
+        -- ^ File patterns whose modification causes an error. Raises an error even if 'shakeLint' is 'Nothing'.
     ,shakeCommandOptions :: [CmdOption]
         -- ^ Defaults to @[]@. Additional options to be passed to all command invocations.
-    ,shakeFlush :: Maybe Double
+    ,shakeFlush :: Maybe Seconds
         -- ^ Defaults to @'Just' 10@. How often to flush Shake metadata files in seconds, or 'Nothing' to never flush explicitly.
         --   It is possible that on abnormal termination (not Haskell exceptions) any rules that completed in the last
         --   'shakeFlush' seconds will be lost.
@@ -147,9 +195,6 @@ data ShakeOptions = ShakeOptions
     ,shakeCreationCheck :: Bool
         -- ^ Default to 'True'. After running a rule to create a file, is it an error if the file does not exist.
         --   Provided for compatibility with @make@ and @ninja@ (which have ugly file creation semantics).
-
---    ,shakeOutputCheck :: Bool
---        -- ^ Default to 'True'. If a file produced by a rule changes, should you rebuild it.
     ,shakeLiveFiles :: [FilePath]
         -- ^ Default to @[]@. After the build system completes, write a list of all files which were /live/ in that run,
         --   i.e. those which Shake checked were valid or rebuilt. Produces best answers if nothing rebuilds.
@@ -157,6 +202,18 @@ data ShakeOptions = ShakeOptions
         -- ^ Defaults to 'False'. Ignore any differences in 'shakeVersion'.
     ,shakeColor :: Bool
         -- ^ Defaults to 'False'. Whether to colorize the output.
+    ,shakeShare :: Maybe FilePath
+        -- ^ Defaults to 'Nothing'. Whether to use and store outputs in a shared directory.
+    ,shakeCloud :: [String]
+        -- ^ Defaults to @[]@. Cloud servers to talk to forming a shared cache.
+    ,shakeSymlink :: Bool
+        -- ^ Defaults to @False@. Use symlinks for 'shakeShare' if they are available.
+        --   If this setting is @True@ (even if symlinks are not available) then files will be
+        --   made read-only to avoid inadvertantly poisoning the shared cache.
+        --   Note the links are actually hard links, not symlinks.
+    ,shakeNeedDirectory :: Bool
+        -- ^ Defaults to @False@. Is depending on a directory an error (default), or it is permitted with
+        --   undefined results. Provided for compatibility with @ninja@.
     ,shakeProgress :: IO Progress -> IO ()
         -- ^ Defaults to no action. A function called when the build starts, allowing progress to be reported.
         --   The function is called on a separate thread, and that thread is killed when the build completes.
@@ -166,6 +223,9 @@ data ShakeOptions = ShakeOptions
         -- ^ Defaults to writing using 'putStrLn'. A function called to output messages from Shake, along with the 'Verbosity' at
         --   which that message should be printed. This function will be called atomically from all other 'shakeOutput' functions.
         --   The 'Verbosity' will always be greater than or higher than 'shakeVerbosity'.
+    ,shakeTrace :: String -> String -> Bool -> IO ()
+        -- ^ Defaults to doing nothing.
+        --   Called for each call of 'Development.Shake.traced', with the key, the command and 'True' for starting, 'False' for stopping.
     ,shakeExtra :: Map.HashMap TypeRep Dynamic
         -- ^ This a map which can be used to store arbitrary extra information that a user may need when writing rules.
         --   The key of each entry must be the 'dynTypeRep' of the value.
@@ -177,52 +237,59 @@ data ShakeOptions = ShakeOptions
 -- | The default set of 'ShakeOptions'.
 shakeOptions :: ShakeOptions
 shakeOptions = ShakeOptions
-    ".shake" 1 "1" Normal False [] Nothing [] [] [] (Just 10) [] [] False True False
-    True ChangeModtime True [] False False
-    (const $ return ())
+    ".shake" 1 "1" Info False [] Nothing [] [] [] [] (Just 10) [] [] False True False
+    True ChangeModtime True [] False False Nothing [] False False
+    (const $ pure ())
     (const $ BS.putStrLn . UTF8.fromString) -- try and output atomically using BS
+    (\_ _ _ -> pure ())
     Map.empty
 
 fieldsShakeOptions =
     ["shakeFiles", "shakeThreads", "shakeVersion", "shakeVerbosity", "shakeStaunch", "shakeReport"
-    ,"shakeLint", "shakeLintInside", "shakeLintIgnore", "shakeCommandOptions"
+    ,"shakeLint", "shakeLintInside", "shakeLintIgnore", "shakeLintWatch", "shakeCommandOptions"
     ,"shakeFlush", "shakeRebuild", "shakeAbbreviations", "shakeStorageLog"
     ,"shakeLineBuffering", "shakeTimings", "shakeRunCommands", "shakeChange", "shakeCreationCheck"
-    ,"shakeLiveFiles","shakeVersionIgnore","shakeProgress", "shakeOutput", "shakeColor", "shakeExtra"]
+    ,"shakeLiveFiles", "shakeVersionIgnore", "shakeColor", "shakeShare", "shakeCloud", "shakeSymlink"
+    ,"shakeNeedDirectory"
+    ,"shakeProgress", "shakeOutput", "shakeTrace", "shakeExtra"]
 tyShakeOptions = mkDataType "Development.Shake.Types.ShakeOptions" [conShakeOptions]
 conShakeOptions = mkConstr tyShakeOptions "ShakeOptions" fieldsShakeOptions Prefix
-unhide x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 y1 y2 y3 =
-    ShakeOptions x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 (fromHidden y1) (fromHidden y2) (fromHidden y3)
+unhide x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 y1 y2 y3 y4 =
+    ShakeOptions x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27
+        (fromHidden y1) (fromHidden y2) (fromHidden y3) (fromHidden y4)
 
 instance Data ShakeOptions where
-    gfoldl k z (ShakeOptions x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 y1 y2 y3) =
+    gfoldl k z (ShakeOptions x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 y1 y2 y3 y4) =
         z unhide `k` x1 `k` x2 `k` x3 `k` x4 `k` x5 `k` x6 `k` x7 `k` x8 `k` x9 `k` x10 `k` x11 `k`
-        x12 `k` x13 `k` x14 `k` x15 `k` x16 `k` x17 `k` x18 `k` x19 `k` x20 `k` x21 `k` x22 `k`
-        Hidden y1 `k` Hidden y2 `k` Hidden y3
-    gunfold k z c = k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ z unhide
+        x12 `k` x13 `k` x14 `k` x15 `k` x16 `k` x17 `k` x18 `k` x19 `k` x20 `k` x21 `k` x22 `k` x23 `k` x24 `k` x25 `k` x26 `k` x27 `k`
+        Hidden y1 `k` Hidden y2 `k` Hidden y3 `k` Hidden y4
+    gunfold k z _ = k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ k $ z unhide
     toConstr ShakeOptions{} = conShakeOptions
     dataTypeOf _ = tyShakeOptions
 
-instance Show ShakeOptions where
-    show x = "ShakeOptions {" ++ intercalate ", " inner ++ "}"
-        where
-            inner = zipWith (\x y -> x ++ " = " ++ y) fieldsShakeOptions $ gmapQ f x
+shakeOptionsFields :: ShakeOptions -> [(String, String)]
+shakeOptionsFields = zipExact fieldsShakeOptions . gmapQ f
+    where
+        f x | Just x <- cast x = show (x :: Int)
+            | Just x <- cast x = show (x :: FilePath)
+            | Just x <- cast x = show (x :: Verbosity)
+            | Just x <- cast x = show (x :: Change)
+            | Just x <- cast x = show (x :: Bool)
+            | Just x <- cast x = show (x :: [FilePath])
+            | Just x <- cast x = show (x :: [(Rebuild, FilePattern)])
+            | Just x <- cast x = show (x :: Maybe Lint)
+            | Just x <- cast x = show (x :: Maybe Double)
+            | Just x <- cast x = show (x :: Maybe String)
+            | Just x <- cast x = show (x :: [(String,String)])
+            | Just x <- cast x = show (x :: Hidden (IO Progress -> IO ()))
+            | Just x <- cast x = show (x :: Hidden (Verbosity -> String -> IO ()))
+            | Just x <- cast x = show (x :: Hidden (Map.HashMap TypeRep Dynamic))
+            | Just x <- cast x = show (x :: Hidden (String -> String -> Bool -> IO ()))
+            | Just x <- cast x = show (x :: [CmdOption])
+            | otherwise = throwImpure $ errorInternal $ "Error while showing ShakeOptions, missing alternative for " ++ show (typeOf x)
 
-            f x | Just x <- cast x = show (x :: Int)
-                | Just x <- cast x = show (x :: FilePath)
-                | Just x <- cast x = show (x :: Verbosity)
-                | Just x <- cast x = show (x :: Change)
-                | Just x <- cast x = show (x :: Bool)
-                | Just x <- cast x = show (x :: [FilePath])
-                | Just x <- cast x = show (x :: [(Rebuild, FilePattern)])
-                | Just x <- cast x = show (x :: Maybe Lint)
-                | Just x <- cast x = show (x :: Maybe Double)
-                | Just x <- cast x = show (x :: [(String,String)])
-                | Just x <- cast x = show (x :: Hidden (IO Progress -> IO ()))
-                | Just x <- cast x = show (x :: Hidden (Verbosity -> String -> IO ()))
-                | Just x <- cast x = show (x :: Hidden (Map.HashMap TypeRep Dynamic))
-                | Just x <- cast x = show (x :: [CmdOption])
-                | otherwise = error $ "Error while showing ShakeOptions, missing alternative for " ++ show (typeOf x)
+instance Show ShakeOptions where
+    show x = "ShakeOptions {" ++ intercalate ", " (map (\(a,b) -> a ++ " = " ++ b) $ shakeOptionsFields x) ++ "}"
 
 
 -- | Internal type, copied from Hide in Uniplate
@@ -232,8 +299,8 @@ newtype Hidden a = Hidden {fromHidden :: a}
 instance Show (Hidden a) where show _ = "<hidden>"
 
 instance Typeable a => Data (Hidden a) where
-    gfoldl k z = z
-    gunfold k z c = error "Development.Shake.Types.ShakeProgress: gunfold not implemented - data type has no constructors"
+    gfoldl _ z = z
+    gunfold _ _ _ = error "Development.Shake.Types.ShakeProgress: gunfold not implemented - data type has no constructors"
     toConstr _ = error "Development.Shake.Types.ShakeProgress: toConstr not implemented - data type has no constructors"
     dataTypeOf _ = tyHidden
 
@@ -242,11 +309,12 @@ tyHidden = mkDataType "Development.Shake.Types.Hidden" []
 
 -- | The verbosity data type, used by 'shakeVerbosity'.
 data Verbosity
-    = Silent -- ^ Don't print any messages.
-    | Quiet  -- ^ Only print essential messages, typically errors.
-    | Normal -- ^ Print errors and @# /command-name/ (for /file-name/)@ when running a 'Development.Shake.traced' command.
-    | Loud   -- ^ Print errors and full command lines when running a 'Development.Shake.command' or 'Development.Shake.cmd' command.
-    | Chatty -- ^ Print errors, full command line and status messages when starting a rule.
+    = Silent  -- ^ Don't print any messages.
+    | Error     -- ^ Only print error messages.
+    | Warn    -- ^ Print errors and warnings.
+    | Info    -- ^ Print errors, warnings and @# /command-name/ (for /file-name/)@ when running a 'Development.Shake.traced' command.
+    | Verbose -- ^ Print errors, warnings, full command lines when running a 'Development.Shake.command' or
+              --   'Development.Shake.cmd' command and status messages when starting a rule.
     | Diagnostic -- ^ Print messages for virtually everything (mostly for debugging).
       deriving (Eq,Ord,Show,Read,Typeable,Data,Enum,Bounded)
 
