@@ -3,7 +3,8 @@ module Development.Shake.Internal.Rules.OrderOnly(
      orderOnly, orderOnlyBS
     ) where
 
-import Development.Shake.Internal.Core.Run
+import Development.Shake.Internal.Core.Types
+import Development.Shake.Internal.Core.Action
 import Development.Shake.Internal.Rules.File
 import qualified Data.ByteString.Char8 as BS
 
