@@ -5,9 +5,10 @@ module Development.Shake.Internal.Rules.Rerun(
     defaultRuleRerun, alwaysRerun
     ) where
 
-import Development.Shake.Internal.Core.Run
 import Development.Shake.Internal.Core.Rules
 import Development.Shake.Internal.Core.Types
+import Development.Shake.Internal.Core.Build
+import Development.Shake.Internal.Core.Action
 import Development.Shake.Classes
 import qualified Data.ByteString as BS
 import General.Binary
@@ -30,14 +31,16 @@ type instance RuleResult AlwaysRerunQ = ()
 --     'Development.Shake.writeFileChanged' out stdout
 -- @
 --
---   In make, the @.PHONY@ attribute on file-producing rules has a similar effect.
+--   In @make@, the @.PHONY@ attribute on file-producing rules has a similar effect.
 --
 --   Note that 'alwaysRerun' is applied when a rule is executed. Modifying an existing rule
 --   to insert 'alwaysRerun' will /not/ cause that rule to rerun next time.
 alwaysRerun :: Action ()
-alwaysRerun = apply1 $ AlwaysRerunQ ()
+alwaysRerun = do
+    historyDisable
+    apply1 $ AlwaysRerunQ ()
 
 defaultRuleRerun :: Rules ()
 defaultRuleRerun =
-    addBuiltinRuleEx noLint $
-        \AlwaysRerunQ{} _ _ -> return $ RunResult ChangedRecomputeDiff BS.empty ()
+    addBuiltinRuleEx noLint noIdentity $
+        \AlwaysRerunQ{} _ _ -> pure $ RunResult ChangedRecomputeDiff BS.empty ()
