@@ -5,8 +5,8 @@ import Development.Shake
 import Development.Shake.FilePath
 import Test.Type
 
-main = shakeTest_ noTest $ do
-    let src = root </> "src/Test/C"
+main = testBuild defaultTest $ do
+    let src = shakeRoot </> "src/Test/C"
     want ["Main.exe"]
 
     "Main.exe" %> \out -> do
@@ -25,4 +25,4 @@ main = shakeTest_ noTest $ do
 cIncludes :: FilePath -> Action [FilePath]
 cIncludes x = do
     Stdout stdout <- cmd "gcc" ["-MM",x]
-    return $ drop 2 $ words stdout
+    pure $ drop 2 $ words stdout
