@@ -4,60 +4,77 @@ module Test.Docs(main) where
 
 import Development.Shake
 import Development.Shake.FilePath
+import qualified System.FilePattern.Directory as IO
 import System.Directory
 import Test.Type
 import Control.Monad
 import Data.Char
-import General.Extra
 import Data.List.Extra
-import Data.Maybe
 import System.Info
 import Data.Version.Extra
 
 
--- Older versions of Haddock garbage the --@ markup and have ambiguity errors
-brokenHaddock = compilerVersion < makeVersion [8]
+-- Older versions of Haddock (GHC 7.10 and below) garbage the --@ markup and have ambiguity errors
+-- GHC 8.0 has a segfault when linking Setup
+brokenHaddock = compilerVersion < makeVersion [8,2]
 
-main = shakeTest_ (unless brokenHaddock . noTest) $ do
+main = testBuild (unless brokenHaddock . defaultTest) $ do
     let index = "dist/doc/html/shake/index.html"
+    let setup = "dist/setup.exe"
     let config = "dist/setup-config"
     want ["Success.txt"]
+    let trackIgnore = trackAllow ["dist/**"]
 
-    let needSource = need =<< getDirectoryFiles "." (map (root </>)
+    let needSource = need =<< getDirectoryFiles "." (map (shakeRoot </>)
             ["src/Development/Shake.hs","src/Development/Shake//*.hs","src/Development/Ninja/*.hs","src/General//*.hs"])
 
+    let runSetup :: [String] -> Action ()
+        runSetup args = do
+            trackIgnore
+            need [setup]
+            -- Make Cabal and Stack play nicely with GHC_PACKAGE_PATH
+            setup <- liftIO $ canonicalizePath setup
+            cmd_ (RemEnv "GHC_PACKAGE_PATH") (Cwd shakeRoot) setup args
+
+    setup %> \_ -> do
+        -- Important to compile the setup binary, or we run foul of
+        -- https://gitlab.haskell.org/ghc/ghc/issues/17575
+        trackIgnore
+        need [shakeRoot </> "Setup.hs"]
+        setup <- liftIO $ canonicalizePath setup
+        curdir <- liftIO $ canonicalizePath "dist"
+        cmd_ (Cwd shakeRoot) "ghc -package=Cabal Setup.hs -o" [setup] "-outputdir" [curdir]
+
     config %> \_ -> do
-        need $ map (root </>) ["shake.cabal","Setup.hs"]
-        -- Make Cabal and Stack play nicely
         path <- getEnv "GHC_PACKAGE_PATH"
-        liftIO $ createDirectoryRecursive "dist"
         dist <- liftIO $ canonicalizePath "dist" -- make sure it works even if we cwd
-        cmd_ (RemEnv "GHC_PACKAGE_PATH") (Cwd root) "runhaskell Setup.hs configure"
-            ["--builddir=" ++ dist,"--user"]
+        need [shakeRoot </> "shake.cabal"]
+        runSetup $
+            ["configure","--builddir=" ++ dist,"--user"] ++
             -- package-db is very sensitive, see #267
             -- note that the reverse ensures the behaviour is consistent between the flags and the env variable
             ["--package-db=" ++ x | x <- maybe [] (reverse . filter (`notElem` [".",""]) . splitSearchPath) path]
+
         -- Paths_shake is only created by "Setup build" (which we want to skip), and required by "Setup haddock", so we fake it
-        copyFile' (root </> "src/Paths.hs") "dist/build/autogen/Paths_shake.hs"
-        copyFile' (root </> "src/Paths.hs") "dist/build/shake/autogen/Paths_shake.hs"
+        copyFile' (shakeRoot </> "src/Paths.hs") "dist/build/autogen/Paths_shake.hs"
+        copyFile' (shakeRoot </> "src/Paths.hs") "dist/build/shake/autogen/Paths_shake.hs"
         writeFile' "dist/build/autogen/cabal_macros.h" ""
         writeFile' "dist/build/shake/autogen/cabal_macros.h" ""
-        trackAllow ["dist//*"]
 
     index %> \_ -> do
-        need $ config : map (root </>) ["shake.cabal","Setup.hs","README.md","CHANGES.txt"]
+        need $ config : map (shakeRoot </>) ["shake.cabal","Setup.hs","README.md","CHANGES.txt","docs/Manual.md","docs/shake-progress.png"]
         needSource
-        trackAllow ["dist//*"]
+        trackIgnore
         dist <- liftIO $ canonicalizePath "dist"
-        cmd (Cwd root) "runhaskell Setup.hs haddock" ["--builddir=" ++ dist]
+        runSetup ["haddock", "--builddir=" ++ dist]
 
     "Part_*.hs" %> \out -> do
-        need [root </> "src/Test/Docs.hs"] -- so much of the generator is in this module
+        need [shakeRoot </> "src/Test/Docs.hs"] -- so much of the generator is in this module
         let noR = filter (/= '\r')
         src <- if "_md" `isSuffixOf` takeBaseName out then
-            fmap (findCodeMarkdown . lines . noR) $ readFile' $ root </> "docs/" ++ drop 5 (reverse (drop 3 $ reverse $ takeBaseName out)) ++ ".md"
+            fmap (findCodeMarkdown . lines . checkBlacklist . noR) $ readFile' $ shakeRoot </> "docs/" ++ drop 5 (reverse (drop 3 $ reverse $ takeBaseName out)) ++ ".md"
          else
-            fmap (findCodeHaddock . noR) $ readFile' $ "dist/doc/html/shake/" ++ replace "_" "-" (drop 5 $ takeBaseName out) ++ ".html"
+            fmap (findCodeHaddock . checkBlacklist . noR) $ readFile' $ "dist/doc/html/shake/" ++ replace "_" "-" (drop 5 $ takeBaseName out) ++ ".html"
 
         let (imports,rest) = partition ("import " `isPrefixOf`) $ showCode src
         writeFileChanged out $ unlines $
@@ -69,7 +86,11 @@ main = shakeTest_ (unless brokenHaddock . noTest) $ do
             ,"import Control.Concurrent"
             ,"import Control.Exception"
             ,"import Control.Monad"
-            ,"import Data.ByteString(ByteString)"
+            ,"import Control.Monad.Trans.Reader"
+            ,"import Data.ByteString(ByteString, pack, unpack)"
+            ,"import qualified Data.ByteString.Char8 as BS"
+            ,"import qualified System.Directory.Extra as IO"
+            ,"import qualified System.IO.Extra as IO"
             ,"import Data.Char"
             ,"import Data.Data"
             ,"import Data.Dynamic"
@@ -77,19 +98,22 @@ main = shakeTest_ (unless brokenHaddock . noTest) $ do
             ,"import System.Time.Extra"
             ,"import Data.Maybe"
             ,"import Data.Monoid"
-            ,"import Development.Shake hiding ((*>),trackAllow)"
+            ,"import Development.Shake hiding ((*>))"
+            ,"import Development.Shake.Command"
             ,"import Development.Shake.Classes"
-            ,"import Development.Shake.Rule hiding (trackAllow)"
+            ,"import Development.Shake.Database"
+            ,"import Development.Shake.Rule"
             ,"import Development.Shake.Util"
             ,"import Development.Shake.FilePath"
             ,"import System.Console.GetOpt"
-            ,"import System.Directory(setCurrentDirectory)"
+            ,"import System.Directory(setCurrentDirectory, withCurrentDirectory)"
             ,"import qualified System.Directory"
-            ,"import System.Environment(lookupEnv, getEnvironment)"
+            ,"import System.Environment(withArgs, lookupEnv, getEnvironment)"
             ,"import System.Process"
             ,"import System.Exit"
             ,"import Control.Applicative"
             ,"import Control.Monad.IO.Class"
+            ,"import Control.Monad.Fail"
             ,"import System.IO"] ++
             ["import " ++ replace "_" "." (drop 5 $ takeBaseName out) | not $ "_md.hs" `isSuffixOf` out] ++
             imports ++
@@ -106,6 +130,7 @@ main = shakeTest_ (unless brokenHaddock . noTest) $ do
             ,"time_elapsed = 1.1"
             ,"old = \"\""
             ,"new = \"\""
+            ,"myvar = \"\""
             ,"myfile = \"\""
             ,"inputs = [\"\"]"
             ,"files = [\"\"]"
@@ -118,7 +143,7 @@ main = shakeTest_ (unless brokenHaddock . noTest) $ do
             ,"instance Eq (OptDescr a)"
             ,"(foo,bar,baz) = undefined"
             ,"(p1,p2) = (0.0, 0.0)"
-            ,"(r1,r2) = (return () :: Rules(), return () :: Rules())"
+            ,"(r1,r2) = (pure () :: Rules(), pure () :: Rules())"
             ,"xs = []"
             ,"ys = []"
             ,"os = [\"file.o\"]"
@@ -127,35 +152,43 @@ main = shakeTest_ (unless brokenHaddock . noTest) $ do
             ,"str2 = \"\""
             ,"def = undefined"
             ,"var = undefined"
+            ,"newValue = undefined"
+            ,"newStore = BS.empty"
+            ,"change = ChangedNothing"
             ,"str = \"\""] ++
             rest
 
     "Files.lst" %> \out -> do
-        need [root </> "src/Test/Docs.hs"] -- so much of the generator is in this module
+        need [shakeRoot </> "src/Test/Docs.hs"] -- so much of the generator is in this module
         need [index]
-        filesHs <- getDirectoryFiles "dist/doc/html/shake" ["Development-*.html"]
-        filesMd <- getDirectoryFiles (root </> "docs") ["*.md"]
+        filesHs <- liftIO $ IO.getDirectoryFiles "dist/doc/html/shake" ["Development-*.html"]
+        -- filesMd on Travis will only include Manual.md, since it's the only one listed in the .cabal
+        -- On AppVeyor, where we build from source, it will check the rest of the website
+        filesMd <- getDirectoryFiles (shakeRoot </> "docs") ["*.md"]
         writeFileChanged out $ unlines $
             ["Part_" ++ replace "-" "_" (takeBaseName x) | x <- filesHs,
                 not $ any (`isSuffixOf` x) ["-Classes.html", "-FilePath.html"]] ++
             ["Part_" ++ takeBaseName x ++ "_md" | x <- filesMd,
-                takeBaseName x `notElem` ["Developing","Model"]]
+                takeBaseName x `notElem` ["Developing","Model","Architecture"]]
 
-    let needModules = do mods <- readFileLines "Files.lst"; need [m <.> "hs" | m <- mods]; return mods
+    let needModules = do mods <- readFileLines "Files.lst"; need [m <.> "hs" | m <- mods]; pure mods
 
     "Main.hs" %> \out -> do
         mods <- needModules
-        writeFileLines out $ ["module Main(main) where"] ++ ["import " ++ m | m <- mods] ++ ["main = return ()"]
+        writeFileLines out $ ["module Main(main) where"] ++ ["import " ++ m ++ "()" | m <- mods] ++ ["main = pure ()"]
 
     "Success.txt" %> \out -> do
-        putNormal . ("Checking documentation for:\n" ++) =<< readFile' "Files.lst"
+        putInfo . ("Checking documentation for:\n" ++) =<< readFile' "Files.lst"
         needModules
         need ["Main.hs"]
-        trackAllow ["dist//*"]
+        trackIgnore
         needSource
-        cmd_ "ghc -fno-code -ignore-package=hashmap " ["-idist/build/autogen","-i" ++ root </> "src","Main.hs"]
+        cmd_ "ghc -fno-code -ignore-package=hashmap" ["-idist/build/autogen","-i" ++ shakeRoot </> "src","Main.hs"]
         writeFile' out ""
 
+checkBlacklist :: String -> String
+checkBlacklist xs = if null bad then xs else error $ show ("Blacklist", bad)
+    where bad = [(w, x) | x <- map lower $ lines xs, w <- blacklist, w `isInfixOf` x]
 
 ---------------------------------------------------------------------
 -- FIND THE CODE
@@ -169,7 +202,8 @@ findCodeHaddock src =
     | tag <- ["code","pre"]
     , x <- insideTag tag src
     , let bad = nubOrd (insideTag "em" x) \\ italics
-    , if null bad then True else error $ "Bad italics, " ++ show bad]
+    , if null bad then True else error $ "Bad italics, " ++ show bad
+    ]
 
 
 findCodeMarkdown :: [String] -> [Code]
@@ -178,9 +212,9 @@ findCodeMarkdown (x:xs) | indented x && not (isBlank x) =
     in Code (dropWhileEnd isBlank $ unindent a) : findCodeMarkdown b
     where
         indented x = length (takeWhile isSpace x) >= 4
-findCodeMarkdown (x:xs) = map (Code . return) (evens $ splitOn "`" x) ++ findCodeMarkdown xs
+findCodeMarkdown (x:xs) = map (Code . pure) (evens $ splitOn "`" x) ++ findCodeMarkdown xs
     where
-        evens (x:y:xs) = y : evens xs
+        evens (_:x:xs) = x : evens xs
         evens _ = []
 findCodeMarkdown [] = []
 
@@ -189,7 +223,7 @@ findCodeMarkdown [] = []
 -- RENDER THE CODE
 
 showCode :: [Code] -> [String]
-showCode = concat . zipWith f [1..] . nubOrd
+showCode = concat . zipWithFrom f 1 . nubOrd
     where
         f i (Code x) | "#" `isPrefixOf` concat x = []
                      | all whitelist x = []
@@ -198,7 +232,7 @@ showCode = concat . zipWith f [1..] . nubOrd
 
 fixCmd :: [String] -> [String]
 fixCmd xs
-    | all ("cmd_ " `isPrefixOf`) xs = xs ++ ["return () :: IO () "]
+    | all ("cmd_ " `isPrefixOf`) xs = xs ++ ["pure () :: IO () "]
     | otherwise = map (replace "Stdout out" "Stdout (out :: String)" . replace "Stderr err" "Stderr (err :: String)") xs
 
 -- | Replace ... with undefined (don't use undefined with cmd; two ...'s should become one replacement)
@@ -208,13 +242,13 @@ undefDots x | Just x <- stripSuffix "..." x, Just (x,_) <- stripInfix "..." x =
     where new = if words x `disjoint` ["cmd","cmd_","Development.Shake.cmd","Development.Shake.cmd_"] then "undefined" else "[\"\"]"
 
 showStmt :: Int -> [String] -> [String]
-showStmt i [] = []
+showStmt _ [] = []
 showStmt i xs | isDecl $ unlines xs = map f xs
     where f x = if fst (word1 x) `elem` dupes then "_" ++ show i ++ "_" ++ x else x
-showStmt i (x:xs) | fst (word1 x) `elem` types = ["type Code_" ++ show i ++ " = " ++ x]
+showStmt i [x] | filter isAlpha (fst $ word1 x) `elem` types = ["type Code_" ++ show i ++ " = " ++ x]
 showStmt i [x] | length (words x) <= 2 = ["code_" ++ show i ++ " = (" ++ x ++ ")"] -- deal with operators and sections
 showStmt i xs | all isPredicate xs, length xs > 1 =
-    zipWith (\j x -> "code_" ++ show i ++ "_" ++ show j ++ " = " ++ x) [1..] xs
+    zipWithFrom (\j x -> "code_" ++ show i ++ "_" ++ show j ++ " = " ++ x) 1 xs
 showStmt i xs = ("code_" ++ show i ++ " = do") : map ("  " ++) xs ++ ["  undefined" | isBindStmt $ last xs]
 
 isPredicate :: String -> Bool
@@ -224,7 +258,7 @@ isBindStmt :: String -> Bool
 isBindStmt x = "let " `isPrefixOf` x || " <- " `isInfixOf` x
 
 isDecl :: String -> Bool
-isDecl x | fst (word1 x) `elem` ["import","infix","instance","newtype"] = True
+isDecl x | fst (word1 x) `elem` ["import","infix","instance","newtype","data"] = True
 isDecl (words -> name:"::":_) | all isAlphaNum name = True -- foo :: Type Signature
 isDecl x | "=" `elem` takeWhile (`notElem` ["let","where"]) (words $ takeWhile (/= '{') x) = True -- foo arg1 arg2 = an implementation
 isDecl _ = False
@@ -248,11 +282,11 @@ dropComment = fst . breakOn "--"
 
 -- | Find all pieces of text inside a given tag
 insideTag :: String -> String -> [String]
-insideTag tag = map (fst . breakOn ("</" ++ tag ++ ">")) . drop 1 . splitOn ("<" ++ tag ++ ">")
+insideTag tag = map (fst . breakOn ("</" ++ tag ++ ">")) . drop1 . splitOn ("<" ++ tag ++ ">")
 
 -- | Given some HTML, find the raw text
 innerText :: String -> String
-innerText ('<':xs) = innerText $ drop 1 $ dropWhile (/= '>') xs
+innerText ('<':xs) = innerText $ drop1 $ dropWhile (/= '>') xs
 innerText ('&':xs)
     | Just xs <- stripPrefix "quot;" xs = '\"' : innerText xs
     | Just xs <- stripPrefix "lt;" xs = '<' : innerText xs
@@ -273,12 +307,12 @@ italics = words "command-name file-name N"
 -- | Identifiers that indicate the fragment is a type
 types :: [String]
 types = words $
-    "MVar IO String FilePath Maybe [String] Char ExitCode Change " ++
+    "MVar IO String FilePath Maybe [String] FSATrace Char ExitCode ReaderT Change " ++
     "Action Resource Rebuild FilePattern Development.Shake.FilePattern " ++
     "Lint Verbosity Rules CmdOption Int Double " ++
     "NFData Binary Hashable Eq Typeable Show Applicative " ++
-    "CmdResult ByteString ProcessHandle Rule Monad Monoid Data TypeRep " ++
-    "BuiltinRun BuiltinLint"
+    "CmdResult ByteString ProcessHandle Rule Monad MonadFail Monoid Data TypeRep " ++
+    "BuiltinRun BuiltinLint BuiltinCheck ShakeDatabase"
 
 -- | Duplicated identifiers which require renaming
 dupes :: [String]
@@ -298,7 +332,7 @@ isCmdFlag x = length a `elem` [1,2] && all (\x -> isAlphaNum x || x `elem` "-=/_
     where (a,b) = span (== '-') x
 
 isCmdFlags :: String -> Bool
-isCmdFlags = all (\x -> let y = fromMaybe x $ stripSuffix "," x in isCmdFlag y || isArg y) . words
+isCmdFlags = all (\x -> let y = dropSuffix "," x in isCmdFlag y || isArg y) . words
     where isArg = all (\x -> isUpper x || x == '=')
 
 isEnvVar :: String -> Bool
@@ -309,25 +343,28 @@ isEnvVar x | Just x <- stripPrefix "$" x = all validChar x
 
 isProgram :: String -> Bool
 isProgram (words -> x:xs) = x `elem` programs && all (\x -> isCmdFlag x || isFilePath x || all isAlpha x || x == "&&") xs
-    where programs = words "excel gcc cl make ghc ghci cabal distcc build tar git fsatrace ninja touch pwd runhaskell rot13 main shake stack rm cat sed sh apt-get build-multiple"
+    where programs = words "excel gcc docker cl make ghc ghci cabal distcc npm build tar git fsatrace ninja touch pwd runhaskell rot13 main shake stack rm cat sed sh apt-get build-multiple"
+isProgram _ = False
 
 -- | Should a fragment be whitelisted and not checked
 whitelist :: String -> Bool
 whitelist x | null x || isFilePath x || isCmdFlags x || isEnvVar x || isProgram x = True
 whitelist x | elem x $ words $
     "newtype do a q m c x value key os contents clean _make " ++
-    ".. /. // \\ //* dir/*/* dir " ++
-    "ConstraintKinds TemplateHaskell GeneralizedNewtypeDeriving DeriveDataTypeable TypeFamilies SetConsoleTitle " ++
+    ".. /. // \\ //* dir/*/* dir [ " ++
+    "ConstraintKinds TemplateHaskell ApplicativeDo OverloadedLists OverloadedStrings GeneralizedNewtypeDeriving DeriveDataTypeable TypeFamilies SetConsoleTitle " ++
     "Data.List System.Directory Development.Shake.FilePath run " ++
     "NoProgress Error src about://tracing " ++
     ".make/i586-linux-gcc/output build " ++
     "/usr/special /usr/special/userbinary " ++
-    "Hidden extension xterm main opts result flagValues argValues " ++
+    "Hidden extension xterm main opts result flagValues argValues fail " ++
     "HEADERS_DIR /path/to/dir CFLAGS let linkFlags temp code out err " ++
     "_shake _shake/build manual chrome://tracing/ compdb " ++
     "docs/manual foo.* _build _build/run depfile 0.000s " ++
     "@ndm_haskell file-name .PHONY filepath trim base stack extra #include " ++
-    "*> BuiltinRun BuiltinLint RuleResult"
+    "*> BuiltinRun BuiltinLint BuiltinIdentity RuleResult " ++
+    "oldStore mode node_modules llbuild Makefile " ++
+    "RebuildNever RLIMIT_NOFILE "
     = True
 whitelist x = x `elem`
     ["[Foo.hi, Foo.o]"
@@ -336,7 +373,7 @@ whitelist x = x `elem`
     ,"1m25s (15%)"
     ,"3m12s (82%)"
     ,"getPkgVersion $ GhcPkgVersion \"shake\""
-    ,"ghc --make MyBuildSystem -threaded -rtsopts \"-with-rtsopts=-I0 -qg -qb\""
+    ,"ghc --make MyBuildSystem -threaded -rtsopts \"-with-rtsopts=-I0 -qg\""
     ,"# command-name (for file-name)"
     ,"<i>build rules</i>"
     ,"<i>actions</i>"
@@ -351,4 +388,19 @@ whitelist x = x `elem`
     ,"$(LitE . StringL . loc_filename <$> location)"
     ,"-d[ FILE], --debug[=FILE]"
     ,"-r[ FILE], --report[=FILE], --profile[=FILE]"
+    ,"man 2 getrlimit"
+    ]
+
+blacklist :: [String]
+blacklist =
+    -- from https://twitter.com/jesstelford/status/992756386160234497
+    ["obviously"
+    ,"basically"
+    ,"simply"
+    ,"of course"
+    ,"clearly"
+    ,"everyone knows"
+    -- ,"however"
+    -- ,"so,"
+    -- ,"easy"
     ]
