@@ -6,14 +6,12 @@ import Development.Shake.Config
 import Development.Shake.FilePath
 import Development.Shake.Util
 import Test.Type
-import Control.Applicative
 import Data.Maybe
-import Prelude
 
 
-main = shakeTest_ noTest $ do
+main = testBuild defaultTest $ do
     -- Example inspired by http://gittup.org/tup/ex_multiple_directories.html
-    usingConfigFile $ root </> "src/Test/Tup/root.cfg"
+    usingConfigFile $ shakeRoot </> "src/Test/Tup/root.cfg"
 
     action $ do
         keys <- getConfigKeys
@@ -24,7 +22,7 @@ main = shakeTest_ noTest $ do
                     | takeExtension x == ".a" = takeBaseName x </> "lib" ++ x
                     | otherwise = error $ "Unknown extension, " ++ x
             x <- fromMaybe (error $ "Missing config key, " ++ key) <$> getConfig key
-            return $ map f $ words x
+            pure $ map f $ words x
 
     (\x -> x -<.> exe == x) ?> \out -> do
         os <- objects "" $ takeBaseName out <.> "exe"
@@ -37,7 +35,7 @@ main = shakeTest_ noTest $ do
         cmd "ar crs" [out] os
 
     "//*.o" %> \out -> do
-        let src = root </> "src/Test/Tup" </> out -<.> "c"
+        let src = shakeRoot </> "src/Test/Tup" </> out -<.> "c"
         need [src]
-        cmd_ "gcc -c -MMD -MF" [out -<.> "d"] [src] "-o" [out] "-O2 -Wall" ["-I" ++ root </> "src/Test/Tup/newmath"]
+        cmd_ "gcc -c -MMD -MF" [out -<.> "d"] [src] "-o" [out] "-O2 -Wall" ["-I" ++ shakeRoot </> "src/Test/Tup/newmath"]
         neededMakefileDependencies $ out -<.> "d"
